#include <cstdio>
#include <math.h>
#ifdef SHAPE
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#endif

#include "card.h"
#include "freecell.h"
#include "option.h"
#include "undo.h"
#include "util.h"

#include "cursor.bm"
#include "cursor_back.bm"
#ifdef SHAPE
#include "boundingMask.bm"
#include "clipMask.bm"
#endif

// Typedef
enum MoveMode { SingleMode, MultipleMode };

// Extern declarations
extern Display* dpy;
extern Window gameWindow;

// Static declarations
#ifdef SHAPE
static bool initialized = false;
static Pixmap boundingMask;
static Pixmap clipMask;
#endif

static char bitmap[bmWidth * (cardHeight - 2)];
static Card* _hilighted = 0;
static bool _cursorChanged = false;
static Cursor _cursor;
static MoveMode moveMode;

Card::Card(Suit s, unsigned int v)
  : NSWindow(true, gameWindow, 0, 0, cardWidth, cardHeight, 1, BlackPixel(dpy, 0))
{
#ifdef SHAPE
  if (Option::roundCard() && !initialized) {
    //Shape
    boundingMask = XCreateBitmapFromData(dpy, root(), boundingMask_bits, 
					 boundingMask_width, boundingMask_height);
    clipMask = XCreateBitmapFromData(dpy, root(), clipMask_bits, clipMask_width, 
				     clipMask_height);

    //Cursor
    XColor fore, back, xc;
    Pixmap p, mask;
    Colormap cm = DefaultColormap(dpy, 0);

    p =  XCreateBitmapFromData(dpy, gameWindow, (char*) cursor_bits,
			       cursor_width, cursor_height);
    mask = XCreateBitmapFromData(dpy, gameWindow, (char*) cursor_back_bits,
				 cursor_back_width, cursor_back_height);
    XAllocNamedColor(dpy, cm, "white", &fore, &xc);
    XAllocNamedColor(dpy, cm, "black", &back, &xc);
    _cursor = XCreatePixmapCursor(dpy, p,mask, &fore, &back, 0, 0);

    initialized = true;
  }
#endif

  _parent = 0;
  _stack = 0;
  _suit = s;
  _value = v;
  _removed = false;
  makeBitmap(s, v, bitmap);

  // initialization on X things
  unsigned long fore, back, hilight;

  if (suitColor(s) == RedSuit)
    fore = getColor(dpy, "red");
  else
    fore = getColor(dpy, "black");
  back = WhitePixel(dpy, 0);
  hilight = getColor(dpy, "lightskyblue4");

  _usualPixmap = 
    XCreatePixmapFromBitmapData(dpy, gameWindow, bitmap, cardWidth - 2,
				cardHeight - 2, fore, back, DefaultDepth(dpy, DefaultScreen(dpy)));
  _hilightedPixmap = 
    XCreatePixmapFromBitmapData(dpy, gameWindow, bitmap, cardWidth - 2,
                                cardHeight - 2, fore, hilight, DefaultDepth(dpy, DefaultScreen(dpy)));

  selectInput(ButtonPressMask | EnterWindowMask | LeaveWindowMask);
  backgroundPixmap(_usualPixmap);
  /*
  XSetWindowAttributes attr;
  long valuemask = CWEventMask | CWBackPixmap | CWBorderPixel;
  attr.event_mask = ButtonPressMask | KeyPressMask | EnterWindowMask | LeaveWindowMask;
  attr.border_pixel = getColor(dpy, "black");
  attr.background_pixmap = _usual;

  Window w;

  w = XCreateWindow(dpy, gameWindow, 0, 0, cardWidth, cardHeight, 1, 
		    (int) CopyFromParent, InputOutput, (Visual *) CopyFromParent,
		    valuemask, &attr);
#ifdef SHAPE
  if (Option::roundCard()) {
    XShapeCombineMask(dpy, w, ShapeBounding, 0, 0, boundingMask, ShapeSet);
    XShapeCombineMask(dpy, w, ShapeClip, 0, 0, clipMask, ShapeSet);
  }
#endif
  
  window(w);
  */

#ifdef SHAPE
  if (Option::roundCard()) {
    XShapeCombineMask(dpy, window(), ShapeBounding, 0, 0, boundingMask, ShapeSet);
    XShapeCombineMask(dpy, window(), ShapeClip, 0, 0, clipMask, ShapeSet);
  }
#endif

  NSWindow::move(0, 0);

  map();
}

void Card::move(int dest_x, int dest_y, bool animate)
{
  if (animate && Option::animation()) {
    int oldx = x();
    int oldy = y();
    int newx = dest_x;
    int newy = dest_y;
    int steps = max(abs(oldx - newx), abs(oldy - newy)) / Option::speedup();
    float curx = (float) oldx;
    float cury = (float) oldy;

    if (steps == 0) {
      NSWindow::move(newx, newy);
      XFlush(dpy);
    } else {
      for (int i = 0; i < steps; i++) {
	curx += ((float) (newx - oldx)) / steps;
	cury += ((float) (newy - oldy)) / steps;
	NSWindow::move((int) curx, (int) cury);
	XFlush(dpy);
      }
    }
  } else {
    NSWindow::move(dest_x, dest_y);
  }

  raise();
}

void Card::moveToStack(Stack* s, bool autoMoving, bool pushUndo)
{
  if (s == 0 || s == _stack) return;

  if (pushUndo)
    undoAddMove(_stack, s);

  _stack->popCard(); 
  _stack = s;

  Card* top = _stack->topCard();
  if (top != 0 && top->canBeParent(this))
    parent(_stack->topCard());
  else
    parent(0);

  _stack->pushCard(this);
  move(_stack->next_x(), _stack->next_y(), true);

  if (autoMoving) autoMove();
}

void Card::moveToStackInitial(Stack* s)
{
  _stack = s;

  Card* top = _stack->topCard();
  if (top != 0 && top->canBeParent(this))
    parent(_stack->topCard());
  else
    parent(0);

  _stack->pushCard(this);
   move(_stack->next_x(), _stack->next_y(), false);
}

bool Card::canBeParent(Card* c) const
{
  if (c == 0) return false;

  return ( suitColor(c->suit()) != suitColor(_suit) ) && (_value == c->value() + 1);
}

void Card::dispatchEvent(const XEvent& ev)
{
  switch (ev.type) {
  case ButtonPress:
    dispatchButtonPress(ev);
    break;
  case EnterNotify:
    dispatchEnterNotify(ev);
    break;
  case LeaveNotify:
    dispatchLeaveNotify(ev);
    break;
  }
}

void Card::dispatchButtonPress(const XEvent& ev)
{
  if (_hilighted == _stack->topCard()) {
    _stack->topCard()->unhilighten();
    _hilighted = 0;
  } else if (_hilighted == 0 && !_removed) { 
    switch (ev.xbutton.button) {
    case 1:
      _stack->topCard()->hilighten();
      _hilighted = _stack->topCard();
      break;
    case 2:
      { 
	SingleStack* stack = emptySingleStack();
	if (stack != 0)
	  _stack->topCard()->moveToStack(stack);
      }
      break;
    }
  } else if (_cursorChanged) {
    // _cursorChanged == true means moving is possible. 
    if (moveMode == SingleMode) {
      _hilighted->unhilighten();
      _hilighted->moveToStack(_stack);
      _hilighted = 0;
      XUndefineCursor(dpy, window());
      _cursorChanged = false;
    } else if (moveMode == MultipleMode) {
      _hilighted->unhilighten();
      moveMultipleCards(_hilighted, _stack->topCard());
      _hilighted = 0;
      XUndefineCursor(dpy, window());
      _cursorChanged = false;
    } else {
      fprintf(stderr, "Bug in Card::dispatchButtonPress _cursorChanged\n");
      exit(1);
    }
  }
}

void Card::dispatchEnterNotify(const XEvent&)
{
  if (_hilighted == 0) return;

  if (_stack->acceptable(_hilighted)) {
    XDefineCursor(dpy, window(), _cursor);
    _cursorChanged = true;
    moveMode = SingleMode;
  } else if (multipleMovable(_hilighted, _stack->topCard())) {
    XDefineCursor(dpy, window(), _cursor);
    _cursorChanged = true;
    moveMode = MultipleMode;
  }
}

void Card::dispatchLeaveNotify(const XEvent&)
{
  if (_cursorChanged) {
    XUndefineCursor(dpy, window());
    _cursorChanged = false;
  }
}

// Interface to External
Card* hilighted()
{
  return _hilighted;
}

void clearHilighted()
{
  if (_hilighted != 0) 
    _hilighted = 0;
}

void unHilighten()
{
  if (_hilighted != 0)
    _hilighted->unhilighten();
}

bool cursorChanged() // There is no bad effect to share variable _cursorChanged.
{
  return _cursorChanged;
}

void setCursor(Window w)
{
  XDefineCursor(dpy, w, _cursor);
  _cursorChanged = true;
}

void clearCursor(Window w)
{
  XUndefineCursor(dpy, w);
  _cursorChanged = false;
}

void clearCursorChanged()
{
  _cursorChanged = false;
}
