#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "card.h"
#include "freecell.h"
#include "general.h"
#include "option.h"
#include "subwindows.h"
#include "undo.h"
#include "util.h"

#include "widget/widget.h"

class NBListener : public NSButtonListener {
public:
  ~NBListener() {}
  void buttonAction(const XEvent&, void *);
};

class UBListener : public NSButtonListener {
public:
  ~UBListener() {}
  void buttonAction(const XEvent&, void*);
};

class SBListener : public NSButtonListener {
public:
  ~SBListener() {}
  void buttonAction(const XEvent&, void *);
};

class PBListener : public NSButtonListener {
public:
  ~PBListener() {}
  void buttonAction(const XEvent&, void *);
};

class EBListener : public NSButtonListener {
public:
  ~EBListener() {}
  void buttonAction(const XEvent&, void *);
};

//Variables declarations
Display* dpy;
Window gameWindow;

static Option* option; // NSInitialize() must be called before this.

// Stacks and cards
static PlayStack* playStack[numPlayStack];
static SingleStack* singleStack[numSingleStack];
static DoneStack* doneStack[numDoneStack];
static DoneStack *doneClub, *doneHeart, *doneDiamond, *doneSpade;

static Card* cards[4][13];

// Buttons
// NSInitialize() must be called before these.
static NSButton* newButton;
static NSButton* undoButton;
static NSButton* scoreButton;
static NSButton* prefButton;
static NSButton* exitButton;

// Listeners for buttons
static NBListener nbl;
static UBListener ubl;
static SBListener sbl;
static PBListener pbl;
static EBListener ebl;

// subwindows
static ScoreWindow* scoreWindow;
static SingleOrMultiple* singleOrMultiple;
static AnotherOrQuit* anotherOrQuit;

const int hGap = 4;
const int vGap = 16;

//Functions declarations
static void adjustSubwindow(NSWindow*);
static bool gameWon();

static void distributeCards(unsigned int = 0);
static void redistributeCards(unsigned int = 0);

static void moveToDoneStackIfPossible(Card*);
static void moveToAppropriateDoneStack(Card*);
static bool canAutoMove(Card*);

static void movePlayStackToPlayStack(Stack*, Stack*, unsigned int);
static unsigned int numEmptySingleStacks();
static unsigned int numEmptyPlayStacks();
static PlayStack* emptyPlayStack();

//Code 
int main(int argc, char* argv[])
{
  Option::parse(argc, argv);

  NSInitialize();
  dpy = NSdisplay();
  option = new Option;
  newButton = new NSButton("New", &nbl);
  undoButton = new NSButton("Undo", &ubl);
  scoreButton = new NSButton("Score", &sbl);
  prefButton = new NSButton("Pref", &pbl);
  exitButton = new NSButton("Exit", &ebl);
  scoreWindow = new ScoreWindow;
  singleOrMultiple = new SingleOrMultiple;
  anotherOrQuit = new AnotherOrQuit;
  
  // toplevel
  Window toplevel;
  unsigned long bgpixel = getColor(dpy, DefaultBackground);
  XSizeHints sh;
  NSStaticHContainer menuContainer;
  NSFrame menuFrame;

  // NSWidget is too powerless to make NSFrame gameWindow.
  toplevel = XCreateSimpleWindow(dpy, RootWindow(dpy, 0), 0, 0, mainWindowWidth,
				 mainWindowHeight, 0, 0, bgpixel);
  XSelectInput(dpy, toplevel, StructureNotifyMask);
  XStoreName(dpy, toplevel, "XFreeCell");
  XSetIconName(dpy, toplevel, "XFreeCell");
  sh.max_width = mainWindowWidth;
  sh.max_height = 1600;
  sh.flags = PMaxSize;
  XSetWMNormalHints(dpy, toplevel, &sh);
  XMapWindow(dpy, toplevel);

  // menu
  menuContainer.add(newButton);
  menuContainer.add(undoButton);
  menuContainer.add(scoreButton);
  menuContainer.add(prefButton);
  menuContainer.add(exitButton);
  menuContainer.reallocate();
  menuFrame.container(&menuContainer);
  menuFrame.parent(toplevel);
  menuFrame.map();

  int menuHeight = menuFrame.height();

  // gameWindow
  gameWindow = XCreateSimpleWindow(dpy, toplevel, 0, menuHeight, mainWindowWidth,
				   mainWindowHeight - menuHeight, 0, 0, bgpixel);
  XMapWindow(dpy, gameWindow);

  adjustSubwindow(option);
  adjustSubwindow(scoreWindow);
  adjustSubwindow(singleOrMultiple);
  adjustSubwindow(anotherOrQuit);
  
  // card
  for (int i = 0; i < 13; i++) 
    cards[0][i] = new Card(Heart, i);
  for (int i = 0; i < 13; i++) 
    cards[1][i] = new Card(Diamond, i);
  for (int i = 0; i < 13; i++) 
    cards[2][i] = new Card(Club, i);
  for (int i = 0; i < 13; i++) 
    cards[3][i] = new Card(Spade, i);

  // stack
  for (unsigned int i = 0; i < numPlayStack; i++)
    playStack[i] = new PlayStack(i * (cardWidth + hGap) + hGap, 2 * vGap + cardHeight);
  for (unsigned int i = 0; i < numSingleStack; i++)
    singleStack[i] = new SingleStack((i + (numPlayStack / 2 - 2)) * (cardWidth + hGap) + hGap, vGap);
  doneStack[0] = new DoneStack(0 * (cardWidth + hGap) + hGap, vGap, Heart);
  doneHeart = doneStack[0];

  doneStack[1] = new DoneStack(1 * (cardWidth + hGap) + hGap, vGap, Spade);
  doneSpade = doneStack[1];

  doneStack[2] = new DoneStack((numPlayStack - 2) * (cardWidth + hGap) + hGap, vGap, Diamond);
  doneDiamond = doneStack[2];

  doneStack[3] = new DoneStack((numPlayStack - 1) * (cardWidth + hGap) + hGap, vGap, Club);
  doneClub = doneStack[3];

  distributeCards();

  while (true) {
    XEvent ev;
    //    NSWindow* win;

    XNextEvent(dpy, &ev);
    NSDispatchEvent(ev);
    //    if ((win = NSWindow::windowToNSWindow(ev.xany.window)) != 0)
    //      win->dispatchEvent(ev);

    if (gameWon()) {
      scoreWindow->incWins();
      anotherOrQuit->waitForEvent();
      if (anotherOrQuit->another())
	redistributeCards();
      else if (anotherOrQuit->quit())
	exit(0);
    }
  }
}

inline void adjustSubwindow(NSWindow* nsw)
{
  nsw->parent(gameWindow);
  nsw->move(mainWindowWidth / 2 - nsw->width() / 2,
	    mainWindowHeight / 2 - nsw->height() / 2);
  nsw->borderWidth(1);
}

inline bool gameWon()
{
  return numEmptyPlayStacks() == numPlayStack;
}

void distributeCards(unsigned int gameNumber)
{
  if (gameNumber == 0) {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    srand((unsigned int) (tv.tv_sec + tv.tv_usec));
    gameNumber = rand();
  }

  // Using the algorithm of MS Freecell
  srand(gameNumber);

  Card* deck[52];

  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 13; j++)
      deck[i * 13 + j] = cards[i][j];

  int wLeft = 52;
  for (int i = 0; i < 52; i++) {
    int j;
    j = rand() % wLeft;
    deck[j]->moveToStackInitial(playStack[(i % numPlayStack)]);
    deck[j] = deck[--wLeft];
  }
}

void redistributeCards(unsigned int gameNumber)
{
  for (unsigned int i = 0; i < 52; i++)
    cards[i / 13][i % 13]->initialize();
  for (unsigned int i = 0; i < numPlayStack; i++)
    playStack[i]->initialize();
  for (unsigned int i = 0; i < numSingleStack; i++)
    singleStack[i]->initialize();
  for (unsigned int i = 0; i < numDoneStack; i++)
    doneStack[i]->initialize();
  clearHilighted();
  clearCursorChanged();
  distributeCards(gameNumber);
}

void NBListener::buttonAction(const XEvent&, void*)
{
  scoreWindow->incDefeats();
  undoClearMoves();
  redistributeCards();
}

void UBListener::buttonAction(const XEvent&, void*)
{
  scoreWindow->incUndos();
  undoDoUndo();
}

void SBListener::buttonAction(const XEvent&, void*)
{
  scoreWindow->waitForEvent();
}

void PBListener::buttonAction(const XEvent&, void*)
{
  option->waitForEvent();
}

void EBListener::buttonAction(const XEvent&, void*)
{
  scoreWindow->incDefeats();
  exit(0);
}

// Functions for automove
void moveToDoneStackIfPossible(Card* card)
{
  if (card == 0) return;

  if (card->value() == 0 || canAutoMove(card)) 
    moveToAppropriateDoneStack(card);
}

void moveToAppropriateDoneStack(Card* card)
{
  switch (card->suit()) {
  case Heart:
    if (doneHeart->acceptable(card))
      card->moveToStack(doneHeart);
    break;
  case Diamond:
    if (doneDiamond->acceptable(card))
      card->moveToStack(doneDiamond);
    break;
  case Spade:
    if (doneSpade->acceptable(card))
      card->moveToStack(doneSpade);
    break;
  case Club:
    if (doneClub->acceptable(card))
      card->moveToStack(doneClub);
    break;
  default:
    fprintf(stderr, "Bug moveToAppropriateDoneStack\n");
    exit(1);
  }
}

bool canAutoMove(Card* card)
{
  if (suitColor(card->suit()) == RedSuit) {
    Card* topSpade = doneSpade->topCard();
    Card* topClub = doneClub->topCard();

    if (topSpade == 0 || topClub == 0)
      return false;
    else if (topClub->value() >= card->value() - 1 &&
             topSpade->value() >= card->value() - 1)
      return true;
    else
      return false;
  } else if (suitColor(card->suit()) == BlackSuit) {
    Card* topHeart = doneHeart->topCard();
    Card* topDiamond = doneDiamond->topCard();

    if (topHeart == 0 || topDiamond == 0)
      return false;
    else if (topHeart->value() >= card->value() - 1 &&
             topDiamond->value() >= card->value() - 1)
      return true;
    else
      return false;
  } else {
    fprintf(stderr, "Bug in canAutoMove\n");
    exit(1);
  }
}

// Functions for multipleMove
void movePlayStackToPlayStack(Stack* from, Stack* to, unsigned int num)
{
  if (num < 0) fprintf(stderr, "Negative num -- movePlayStackToPlayStack\n");
  if (from->size() < num) num = from->size();

  unsigned int numEmptySingleStack = numEmptySingleStacks();
  SingleStack* sStack;
  Card* movedCards[numSingleStack];
  unsigned int index = 0;

  if (numEmptySingleStack + 1 < num) num = numEmptySingleStack + 1;

  Card* tmp;
  while (index < numEmptySingleStack && num > 0) {
    sStack = emptySingleStack();
    tmp = from->topCard();
    tmp->moveToStack(sStack);
    movedCards[index++] = tmp;
    num--;
  }

  if (num != 0)
    from->topCard()->moveToStack(to);

  for (; index > 0; index--) {
    if (!movedCards[index - 1]->isRemoved())
      movedCards[index - 1]->moveToStack(to);
  }
}

// Util Functions
unsigned int numEmptySingleStacks()
{
  unsigned int num = 0;

  for (unsigned int i = 0; i < numSingleStack; i++)
    if (singleStack[i]->size() == 0) num++;

  return num;
}

unsigned int numEmptyPlayStacks()
{
  unsigned int num = 0;
  
  for (unsigned int i = 0; i < numPlayStack; i++)
    if (playStack[i]->size() == 0) num++;

  return num;
}

PlayStack* emptyPlayStack()
{
  for (unsigned int i = 0; i < numPlayStack; i++) {
    if (playStack[i]->size() == 0)
      return playStack[i];
  }

  return 0;
}  

// Interface to external
void autoMove()
{
  Card* card;
  for (unsigned int i = 0; i < numPlayStack; i++) {
    card = playStack[i]->topCard();
    moveToDoneStackIfPossible(card);
  }
  for (unsigned int i = 0; i < numSingleStack; i++) {
    card = singleStack[i]->topCard();
    moveToDoneStackIfPossible(card);
  }
}

SingleStack* emptySingleStack()
{
  for (unsigned int i = 0; i < numSingleStack; i++) {
    if (singleStack[i]->size() == 0)
      return singleStack[i];
  }

  return 0;
}

bool multipleMovable(Card* from, Card* to)
{
  if (from == 0 || to == 0) return false;

  // Checks whether from's ancestors are accpeted
  Stack* destStack = to->stack();
  unsigned int movableNum = 1;
  bool found = false;

  for (Card* c = from; c != 0; c = c->parent(), movableNum++) {
    if (destStack->acceptable(c)) {
      found = true;
      break;
    }
  }

  if (!found) return false;

  // Checks whether there are enough spaces
  unsigned int numEmptySS = numEmptySingleStacks();
  unsigned int numEmptyPS = numEmptyPlayStacks();

  return movableNum <= (numEmptySS + 1) * (numEmptyPS + 1);
}

void moveMultipleCards(Card* from, Card* to)
{
  Stack* fromStack = from->stack();
  Stack* toStack = to->stack();

  int numCardsToBeMoved = to->value() - from->value();
  int numMovableCards = numEmptySingleStacks() + 1;
  int index = 0;
  PlayStack* tempStack[numPlayStack];
  int numEmptyPlayStack = numEmptyPlayStacks();
  
  PlayStack* tmp;
  for (; numEmptyPlayStack > 0 && numCardsToBeMoved > numMovableCards;
       numEmptyPlayStack--, numCardsToBeMoved -= numMovableCards) {
    tmp = emptyPlayStack();
    movePlayStackToPlayStack(fromStack, tmp, numMovableCards);
    tempStack[index++] = tmp;
  }

  if (numCardsToBeMoved > 1)
    movePlayStackToPlayStack(fromStack, toStack, numCardsToBeMoved);

  if (numCardsToBeMoved == 1)
    fromStack->topCard()->moveToStack(toStack);

  for (; index > 0; index--)
    movePlayStackToPlayStack(tempStack[index - 1], toStack, numMovableCards);
}

void moveMultipleCards(Card* from, PlayStack* destStack)
{
  Stack* fromStack = from->stack();
  
  unsigned int numMovableCards = 0;
  for (; from != 0; from = from->parent(), numMovableCards++)
    ;

  movePlayStackToPlayStack(fromStack, destStack, numMovableCards);
}
